require 'thor'
require 'json'
require 'time'
require 'colorize'
require 'socket'
require 'net/http'
require 'net/https'
require 'uri'
require 'openssl'
require 'timeout'
require "highline/import"

module Echo_cli
  class Echo < Thor
    desc "post", "Post metrics to Echo. Use \"echo_cli post '<metric>' {number} {-z}\""

    option :z

    long_desc <<-LONGDESC
      Post metrics to Echo. Use \"echo_cli post '<metric>' {number} {-z}\"

      `echo_cli post` will post a metric directly to Echo's StatsD server.

      You can optionally specify a third integer parameter, which will post the same metric an integer number of times equal to the number specified.

      Ensure single-quotes surround the metric.

      Ensure the environment variable 'ECHO_CLI_HOST' is set to the correct IP.

      With the -z (zombie) option, post will continuously send the metric to Echo a random number of times every second in the background, and return its PID for future removal. Use "$ kill -9 <PID>" to kill the zombie process.

      The -z option will override the {number} parameter, should both be included in an execution of the command.

      > $ echo_cli post 'example.post:1|c' {10} {-z}
    LONGDESC

    def post(metric, num = 1)
      helper = Helper.new
      helper.set_envs()
      helper.check_for_IP()

      metric = metric.gsub(/'/,"").gsub(/"/,"")
      exit if !helper.check_metric_format(metric)

      if options[:z]
        time_start = Time.new.to_i.to_s

        # sort of a mini-smoketest to ensure the posts will work before creating a background process
        helper.do_post_udp(ENV["ECHO_CLI_HOST"], "8125", metric)

        pid = Process.fork do
          prng = Random.new
          while true do
            helper.do_post_udp(ENV["ECHO_CLI_HOST"], "8125", metric)
            sleep prng.rand(0.1...1.0)
          end
        end
        puts "\nProcess ID: ".green + pid.to_s.yellow + "\nUse ".green + "$ kill ".yellow + pid.to_s.yellow + " to kill the zombie process.".green
        puts "Process began posting at ".green + time_start.yellow
      else
        num = num.to_i
        if num <= 0
          puts "\nThe number of posts specified is an incorrect expression, or is less than or equal to zero!".red
          exit
        end

        while num > 0
          helper.do_post_udp(ENV["ECHO_CLI_HOST"], "8125", metric)
          puts "\nPosted ".green + metric.yellow + " to ".green + "http://".yellow + ENV["ECHO_CLI_HOST"].yellow + ":".yellow + "8125".yellow + " at ".green + Time.new.to_i.to_s.yellow
          num -= 1
        end
      end

      helper.unset_envs()
    end

    desc "query", "Query metrics on Echo. Use \"echo_cli query '<metric>' {time_start} {time_end} {-n}\""

    option :n

    long_desc <<-LONGDESC
      Query metrics on Echo. Use \"echo_cli query '<metric>' {time_start} {time_end} {-n}\"

      `echo_cli query` will query a metric directly from Echo's Graphite server instance.

      The default timestamp query range begins 15 minutes prior to the current time, to now.

      See: https://graphite-api.readthedocs.io/en/latest/api.html#from-until for possible time_start and time_end formats.

      The metric-value number (following the ':' in the <metric> string parameter) can be random. The query command does not require this information, it is merely left in place for consistency across commands and ease-of-use.

      Ensure single-quotes surround the metric.

      Ensure the environment variable 'ECHO_CLI_HOST' is set to the correct IP.

      By default, the CLI removes any null datapoints from the returned query, so an empty dataset returned upon a successful query would suggest that either not datapoints were found in the query range, or that the metric was spelled incorrectly.

      The -n option can be used to print all of the datapoints - including null datapoints - resulting from a query.

      > $ echo_cli query 'example.query:1|c' {1492186212} {-5min}
    LONGDESC

    def query(metric, time_start = '-15min', time_end = 'now')
      helper = Helper.new
      helper.set_envs()
      helper.check_for_IP()
      metric = metric.gsub(/'/,"").gsub(/"/,"")
      exit if !helper.check_metric_format(metric)

      query_url = 'http://' + ENV["ECHO_CLI_HOST"] + ':8080/render?target=' + helper.get_metric_string(metric) + '&from=' + time_start + '&until=' + time_end + '&format=json'

      response = helper.query_graphite(query_url)
      if(response.code != '200')
        puts "\nThe query returned a status code of ".red + response.code.yellow + ". If the query has start or end times, ensure they are formatted correctly.".red
        puts "Otherwise, check your metric query formatting and the 'ECHO_CLI_HOST' environment variable value.".red
        puts "\nResponse Body:\n".red
        puts response.body
      else
        res = JSON.parse(response.body)
        if(!options[:n])
          res.each_with_index{|target, index| res[index]["datapoints"].delete_if{|value| value[0] == nil}}
        end
        puts "\nSuccessful Query:\n\n".green + JSON.pretty_generate(res).to_s.green
      end
      helper.unset_envs()
    end

    desc "populate", "Back-populate a metric. Use \"echo_cli populate '<metric>' {timespan} {frequency}\""

    long_desc <<-LONGDESC
      Back-populate a metric for a specified amount of time. Use \"echo_cli populate '<metric>' {timespan} {frequency}\"

      `echo_cli populate` will populate a metric's timeseries data with random values between 1-100.

      The default {timespan} that will be populated spans from now to 1209600 seconds (2 weeks) prior to now.

      The default {frequency} at which data will be populated is 600 seconds (10 minutes).

      The metric-value number (following the ':' in the <metric> string parameter) can be random. The populate command does not require this information, it is merely left in place for consistency across commands and ease-of-use.

      Ensure single-quotes surround the metric.

      Ensure the environment variable 'ECHO_CLI_HOST' is set to the correct IP.

      > $ echo_cli populate 'example.back.populate:1|c' {600} {100}
    LONGDESC

    def populate(metric, timespan = 1209600, frequency = 600)
      helper = Helper.new
      helper.set_envs
      metric = metric.gsub(/'/,"").gsub(/"/,"")
      exit if !helper.check_metric_format(metric)

      now = Time.new.to_i
      prng = Random.new

      puts "\nWARNING: You are about to back-populate a metric.\nAre you sure you want to back-populate ".red + metric.yellow + " for the last ".red + timespan.to_s.yellow + " seconds?".red
      input = ask "[y/n]".red

      if input == "y"
        puts "\nPopulating..."
        timespan = timespan.to_i
        frequency = frequency.to_i
        while timespan >= 0
          to_post = helper.get_metric_string_populate(metric) + " " + prng.rand(0...100).to_s + " " + (now - timespan).to_s
          helper.do_post_tcp(ENV["ECHO_CLI_HOST"], "2003", to_post)
          timespan = timespan - frequency
        end
        puts "Complete."
      else
        exit
      end
    end

    desc "convert", "Convert timestamp formats. Use \"echo_cli convert <timestamp>\""

    long_desc <<-LONGDESC
      Convert timestamp formats. Use \"echo_cli convert <timestamp>\"

      `echo_cli convert` will convert a given timestamp from epoch to a human-readable format, or vice-versa. Useful for determining locations of data on grafana. Ensure that an epoch timestamp has no punctation, and a human-readable timestamp follows the format 'YYYY-MM-DDTHH:MM:SS.SSS-Z' exactly.

      Epoch timestamp example: 1492186214

      Human-readable date/time timestamp example: 2011-10-18T21:01:52-07:00

      > $ echo_cli convert 1492186214

      > $ echo_cli convert 2011-10-18T21:01:52-07:00
    LONGDESC

    def convert(timestamp)
      if timestamp.include? ":"
        time = Time.iso8601(timestamp).to_i.to_s
      else
        time = Time.at(timestamp.to_i).to_datetime.to_s
      end
      puts "\nConverted timstamp: ".green + time.yellow
      return time
    end

    desc "now", "The current time, in epoch and human-readable formats. Use \"echo_cli now\""

    long_desc <<-LONGDESC
      The current time, in epoch and human-readable formats. Use \"echo_cli now\"

      `echo_cli now` will present the current timestamp (based on your computer's local time) to more easily utilize the rest of the Echo CLI commands.

      Epoch timestamp example: 1492186214

      Human-readable date/time timestamp example: 2011-10-18T21:01:52-07:00

      > $ echo_cli now
    LONGDESC

    def now
      timestamp = Time.new.to_i.to_s
      puts "\nCurrent Epoch time: ".green + timestamp.to_s.yellow + "\nCurrent time: ".green + Time.at(timestamp.to_i).to_datetime.to_s.yellow
    end
  end

  class Helper
    # useful silencing function, plz leave here
    def stfu
      begin
        orig_stderr = $stderr.clone
        orig_stdout = $stdout.clone
        $stderr.reopen File.new('/dev/null', 'w')
        $stdout.reopen File.new('/dev/null', 'w')
        retval = yield
      rescue Exception => e
        $stdout.reopen orig_stdout
        $stderr.reopen orig_stderr
        raise e
      ensure
        $stdout.reopen orig_stdout
        $stderr.reopen orig_stderr
      end
      retval
    end

    def set_envs
      ENV["http_proxy"] = ENV["HTTP_PROXY"]
      ENV["https_proxy"] = ENV["HTTPS_PROXY"]
    end

    def unset_envs
      ENV["http_proxy"] = ENV[""]
      ENV["https_proxy"] = ENV[""]
    end

    def check_for_IP
      if !ENV["ECHO_CLI_HOST"] || ENV["ECHO_CLI_HOST"] == ''
        puts "\nFailed to find IP or URL. Please set the environment variable: ".red + "'ECHO_CLI_HOST'".yellow + " to continue.".red
        exit
      end
    end

    def get_metric_string(metric)
      type = metric[metric.index('|') + 1 .. metric.length]
      case type
      when 'c'
        return metric[0..metric.index(':')-1] + ".*"
      when 'ms'
        return metric[0..metric.index(':')-1] + ".*"
      when 's'
        return metric[0..metric.index(':')-1] + ".*"
      when 'g'
        return metric[0..metric.index(':')-1]
      end
    end

    def get_metric_string_populate(metric)
      type = metric[metric.index('|') + 1 .. metric.length]
      case type
      when 'c'
        return metric[0..metric.index(':')-1] + ".count"
      when 'ms'
        return metric[0..metric.index(':')-1] + ".count"
      when 's'
        return metric[0..metric.index(':')-1] + ".count"
      when 'g'
        return metric[0..metric.index(':')-1]
      end
    end

    def check_metric_format(metric)
      if /.*(:\d*\.?\d*)\|(ms|c|g|s)/.match(metric).to_s != metric
        puts "\nThe posted metric was not of the correct form.\nMetrics must be of the form ".red + "<Application_name>.<tag1>.<tag2>...<tag(n)>:<value>|<metric_type>".yellow
        puts "Where <value> is a number, and <type> is any of the following: 'ms', 'c', 's', 'g'.".red
        return false
      end
      return true
    end

    def do_post_udp(statsd_uri, port, request_body)
      begin
        udpsocket = UDPSocket.new
        udpsocket.connect(statsd_uri, port)
        udpsocket.write(request_body)
        udpsocket.close
      rescue Errno::ECONNREFUSED
      end
    end

    def do_post_tcp(statsd_uri, port, request_body)
      begin
        timeout(10) do
          socket = TCPSocket.new(statsd_uri, port)
          socket.write(request_body + "\n")
          socket.close
        end
      rescue Timeout::Error
        puts "\nOperation timed out! Ensure the 'ECHO_CLI_HOST' environment variable is correct.".red
        exit
      rescue Errno::ECONNREFUSED
      end
    end

    def query_graphite(graphite_uri)
      uri = URI.parse(graphite_uri)
      https = Net::HTTP.new(uri.host, uri.port)
      https.use_ssl = false
      https.verify_mode = OpenSSL::SSL::VERIFY_NONE if https.use_ssl?
      req = Net::HTTP::Get.new(uri.request_uri, initheader = {'Content-Type' => 'text/plain'}, )
      res = https.request(req)
      return res
    end
  end
end
